/**
 * @file   core1553brm_as250.c
 * @ingroup core1553brm_as250
 * @author Armin Luntzer (armin.luntzer@univie.ac.at),
 * @date   July, 2015
 *
 * @copyright GPLv2
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * @brief Driver for the MicroSemi Core 1553 BRM implementing the AS250
 *        protocol layer
 *
 * @defgroup core1553brm_as250 Core1553BRM/AS250 device/protocol driver
 *
 *
 *
 * ## Overview
 *
 * This is a driver for the MicroSemi Core1553BRM implementing the
 * __AS250 MIL-STD-1553__ Bus Protocol as specified in
 * __DIV.SP.00030.T.ASTR__ _Issue 1 Revision 1_, tailored as specified in
 * __CHEOPS-ECE-CSW-ICD-049__ _Issue 4.x_.
 *
 * The AS250 1553 protocol specification is based on the __ECSS-E-ST-50-13C__
 * standard.
 *
 * The main differences in this protocol variant are:
 *
 * - twice as many Acquisition Transfer Requests (ATR) are allowed for the
 *   Remote Terminal (RT) per major frame
 * - the doubled ATR rate results in an increased transfer volume of
 *   128kbits per second
 * - no confirmations of ATRs are written by the Bus Controller
 * - the transfer block counter wraps from 255 to 0 instead of 1
 * - a transfer will be accepted as new as long as the block counter differs
 *   from the previous one
 *
 *
 * ## Mode of Operation
 *
 * @see _Core1553BRM v4.1 Handbook_ for a description of the hardware
 * interface
 *
 * Since the protocol requires an response to a data write by the
 * _Bus Controller_ (the spacecraft) in the next 1553 minor frame, the least
 * involved approach is to operate the 1553 core in _ping-pong_ mode.
 * This mode is a double-buffering scheme, in which two buffers are assigned
 * per RX and TX subaddress. As the core operates, it will use one of the
 * buffers for ongoing transfers, while the other is available to the software
 * for reading or writing. In the next minor frame, the buffers will be
 * switched and so on.
 *
 * The interaction with the 1553 core is triggered by interrupts that are
 * generated by the former based on its subaddress configuration. The attached
 * interrupt service routine (ISR) of the driver then inspects the interrupt log
 * list of the 1553 core and processes recorded events. If a transfer occured,
 * it reads or writes data from/to the subaddress buffers and handles the
 * protocol response. Telecommands (_Distribution Data_ or __DD__) written by
 * the spacecraft are stored in @ref packet_tracker buffers that are queried by
 * the application software. Telemetry data (_Acquisition Data_ or __AD__) are
 * extracted from a @ref cpus_buffer that accepts at most 2 kiB of PUS packets
 * from the IASW at any given time.
 *
 * @image html core1553brm_as250/irq_handler.png "Core1553BRM driver ISR"
 *
 *
 * ### Acquisition Packets
 *
 * The IASW is by design not able to supply prefabricated transfer frames of
 * telemetry packets to the driver directly. Instead, it tries to hand off as
 * many of packets as possible in one polling cycle. Therefore, the IBSW
 * therefore supplies the @ref cpus_buffer that acts as an intermediate to the
 * 1553 driver by properly packing TM data into transfer frames that do not
 * split PUS packets across a minor frame transfer, as required by the
 * communication protocol.
 *
 * Aquisition Transfers are issued every time TX subaddress (__SA__) 28 is read
 * by the Bus Controller. The inactive ping-pong buffer is determined and a
 * frame is constructed and loaded into TX subaddresses 11-26, then the block
 * counter in the Acquisition Transfer Request message is updated and the latter
 * loaded into SA 28.
 *
 * @startuml {AT_generation.svg} "Acquisition Transfer Generation" width=10
 *
 * start
 *
 * :get last ATR;
 *
 * :create transfer frame;
 *
 * if (no packet) then (yes)
 *
 * while (packet fits?)
 *	:add packet;
 * endwhile
 *
 * :update block counter;
 *
 * while (bytes left?)
 *	:write SA;
 *	:next SA;
 * endwhile
 *
 * :write ATR;
 *
 * endif
 * stop
 *
 * @enduml
 *
 *
 * ### Distribution Packets
 *
 * Telecommand packets written by the Bus Controller are directed into different
 * buffers by inspecting their PUS header for their source ID and stored in two
 * @ref packet_tracker buffers. This has to be done in the driver, because the
 * the IASW cannot determine the source of a packet itself. Instead, it requires
 * an interface supplied by the IBSW for each packet sources it allows.
 *
 * @image html core1553brm_as250/split_packetstreams_direct_buffer_drop.png "Packet redirection into buffers based on ID"
 *
 * ## Error Handling
 *
 * Most errors reported by the 1553 core during run time are either non-critical
 * or usually non-recoverable. The latter usually result from breakage on the
 * physical layer of the 1553 bus, or from severe misconfiguration of the core.
 * In either case, no reports on the state of the device can be expected to
 * reach the bus controller.
 *
 * Errors that may have been raised are reported via the event reporting
 * interface. If a critical errors warranting a FDIR response occured, an
 * optional, user-supplied reset function is called, which can for example
 * result in entering safe mode, where there is a chance to handle
 * misconfiguration via the boot software.
 *
 * Errors in the @ref cpus_buffer or @ref packet_tracker are not critical to
 * core operation and result in a full reset of the buffers issued by the driver
 * itself, accompanied by an error report.
 *
 *
 * ## Notes
 *
 * The implementation of the protocol in the driver layer was necessitated due
 * to the low-frequency polling-style operating of the application software,
 * which would have been too slow to operate the exchange by itself.
 * A separate dedicated thread was not an option, because of the FIFO
 * scheduling policy selected for the IFSW and also the requirement to allow
 * packet priorisation within the IASW.
 *
 *
 *
 * @example demo1553_as250.c
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdint.h>

#include <errno.h>

#include <core1553brm_as250.h>
#include <irq.h>
#include <irq_dispatch.h>

#include <cpus_buffers.h>
#include <packet_tracker.h>

#include <errors.h>
#include <compiler.h>
#include <io.h>

#include <event_report.h>
#include <sysctl.h>


/*
 * NOTE: we assume untranslated addresses for now, so there is no need for
 *	 alternate space loads. This needs to be added if the MMU and/or
 *	 priviledged memory access is going to be used with the driver
 */





static struct {
	uint32_t rx_bytes;
	uint32_t tx_bytes;
	uint32_t bytes_dropped;
} brmstat;

#if (__sparc__)
#define UINT32_T_FORMAT		"%lu"
#else
#define UINT32_T_FORMAT		"%u"
#endif

static ssize_t rxtx_show(__attribute__((unused)) struct sysobj *sobj,
			 __attribute__((unused)) struct sobj_attribute *sattr,
			 char *buf)
{
	if (!strcmp(sattr->name, "rx_bytes"))
		return sprintf(buf, UINT32_T_FORMAT, brmstat.rx_bytes);

	if (!strcmp(sattr->name, "tx_bytes"))
		return sprintf(buf, UINT32_T_FORMAT, brmstat.tx_bytes);

	if (!strcmp(sattr->name, "bytes_dropped"))
		return sprintf(buf, UINT32_T_FORMAT, brmstat.bytes_dropped);

	return 0;
}

static ssize_t rxtx_store(__attribute__((unused)) struct sysobj *sobj,
			  __attribute__((unused)) struct sobj_attribute *sattr,
			  __attribute__((unused)) const char *buf,
			  __attribute__((unused)) size_t len)
{
	if (!strcmp(sattr->name, "rx_bytes")) {
		brmstat.rx_bytes = 0;
		return 0;
	}

	if (!strcmp(sattr->name, "tx_bytes")) {
		brmstat.tx_bytes = 0;
		return 0;
	}

	if (!strcmp(sattr->name, "bytes_dropped")) {
		brmstat.bytes_dropped = 0;
		return 0;
	}

	return 0;
}

__extension__
static struct sobj_attribute rx_bytes_attr = __ATTR(rx_bytes,
						    rxtx_show,
						    rxtx_store);
__extension__
static struct sobj_attribute tx_bytes_attr = __ATTR(tx_bytes,
						    rxtx_show,
						    rxtx_store);
__extension__
static struct sobj_attribute dropped_bytes_attr = __ATTR(bytes_dropped,
							 rxtx_show,
							 rxtx_store);

__extension__
static struct sobj_attribute *brm_attributes[] = {&rx_bytes_attr,
						  &tx_bytes_attr,
						  &dropped_bytes_attr,
						  NULL};




/**
 * @brief central driver error handling
 * @param brm a struct brm_config
 * @param s the error severity
 */

static void brm_handle_error(struct brm_config *brm, enum error_severity s)
{
	event_report(CORE1553BRM, s, (uint32_t) errno);

	if (s == HIGH)
		if (brm->rt_reset)
			brm->rt_reset(brm->rt_reset_userdata);
}


/**
 * @brief computes the parity bit for the given RT address
 * @retval 1 : for multiple-of-two numbers of high bits
 * @retval 0 : otherwise
 */

static uint8_t rt_parity_bit(uint8_t rt_address)
{
	/* 1) make sure only valid RTA bits are selected
	 * 2) xor upper nibble with lower nibble to eliminate pairs
	 * 3) take the lower nibble only, it contains all non-pairs in the byte
	 * 4) determine parity from precalculated bit field suitable for the
	 *   RTPTY flag. For dumbos:
	 *   1 = 1 bit, 2 = 1 bit, 3 = 2 bits, 4 = 1 bit, 5 = 2 bits etc...
	 */

	rt_address &= BRM_RTA_BITS;
	rt_address ^= (rt_address >> 4);
	rt_address &= 0x0F;

	return ((0x9669 >> rt_address) & 0x1);
}


/**
 * @brief queries the MSGTO bit to determine whether ping-pong mode is
 *        disabled or enabled
 * @param brm a struct brm_config
 * @retval 1 : enabled
 * @retval 0 : disabled
 */

static int32_t brm_pp_enabled(struct brm_config *brm)
{
	return ioread32be(&brm->regs->ctrl) & BRM_CCR_MSGTO;
}


/**
 * @brief disable ping-pong mode
 * @param brm a struct brm_config
 * @note success of operation should always be checked via brm_pp_enabled()
 * @see Core1553BRM v4.0 Handbook, p.61
 */

static void brm_disable_pp(struct brm_config *brm)
{
	uint32_t ctrl;

	ctrl = ioread32be(&brm->regs->ctrl) & ~BRM_CCR_PPEN;

	iowrite32be(ctrl, &brm->regs->ctrl);
}


/**
 * @brief enable ping-pong mode
 * @param brm a struct brm_config
 * @note success of operation should always be checked via brm_pp_enabled()
 * @see Core1553BRM v4.0 Handbook, p.61
 *
 */

static void brm_enable_pp(struct brm_config *brm)
{
	uint32_t ctrl;

	ctrl = ioread32be(&brm->regs->ctrl) | BRM_CCR_PPEN;

	iowrite32be(ctrl, &brm->regs->ctrl);
}


/**
 * @brief gets offset to currently non-active (secondary) ping-pong data buffer
 *        of a RX subaddress
 * @param brm a struct brm_config
 * @param sa a subaddress number
 * @returns offset into the buffer
 */

static uint32_t brm_get_offset_pp_rx_prev(struct brm_config *brm, uint32_t sa)
{
	uint32_t offset;

	if (ioread16be(&brm->rt_mem->sa_rx[sa].CW) & BRM_RT_CW_FLAGS_AB)
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_rx[sa].buf_A);
	else
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_rx[sa].buf_B);


	return (offset + BRM_HEADER_WORDS);
}


/**
 * @brief gets offset to currently non-active (seconardy) ping-pong data buffer
 *        of a TX subaddress
 * @param brm a struct brm_config
 * @param sa a subaddress number
 * @returns offset into the buffer
 * @note this buffer will be read by the core next when it is polled by the bus
 *       controller
 */

static uint32_t brm_get_offset_pp_tx_next(struct brm_config *brm, uint32_t sa)
{
	uint32_t offset;

	if (ioread16be(&brm->rt_mem->sa_tx[sa].CW) & BRM_RT_CW_FLAGS_AB)
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_tx[sa].buf_B);
	else
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_tx[sa].buf_A);

	return (offset + BRM_HEADER_WORDS);
}


/**
 * @brief gets offset to currently active (primary) ping-pong data buffer
 *        of a TX subaddress
 * @param brm a struct brm_config
 * @param sa a subaddress number
 * @returns offset into the buffer
 * @note this buffer is currently marked for reading by the core
 */

static uint32_t brm_get_offset_pp_tx_curr(struct brm_config *brm, uint32_t sa)
{
	uint32_t offset;

	if (ioread16be(&brm->rt_mem->sa_tx[sa].CW) & BRM_RT_CW_FLAGS_AB)
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_tx[sa].buf_A);
	else
		offset = (uint32_t) ioread16be(&brm->rt_mem->sa_tx[sa].buf_B);

	return (offset + BRM_HEADER_WORDS);
}


/*
 * @brief gets offset to currently non-active (secondary) ping-pong data buffer
 *        of a mode code RX subaddress
 * @param brm a struct brm_config
 * @param sa a subaddress number
 * @returns offset into the buffer
 */

__attribute__((unused))
static uint32_t brm_get_offset_pp_mc_rx_prev(struct brm_config *brm,
					     uint32_t sa)
{
	uint32_t offset;

	if (ioread16be(&brm->rt_mem->mc_rx[sa].CW) & BRM_RT_CW_FLAGS_AB)
		offset = (uint32_t) ioread16be(&brm->rt_mem->mc_rx[sa].buf_A);
	 else
		offset = (uint32_t) ioread16be(&brm->rt_mem->mc_rx[sa].buf_B);


	return (offset + BRM_HEADER_WORDS);
}


/**
 * @brief gets number of data words in a descriptor
 * @param brm a struct brm_config
 * @param descriptor a descriptor number
 * @returns data word count
 * @note the Message Information Word and Time Tag entries are not included in
 *       the count; this is only sensible to use with RX subaddresses;
 */

static uint32_t brm_get_wordcount(struct brm_config *brm, uint32_t descriptor)
{
	uint32_t wc;
	uint32_t offset;

	if (ioread16be(&brm->rt_mem->descriptors[descriptor].CW) &
		 BRM_RT_CW_FLAGS_AB)
		offset = (uint32_t)
			ioread16be(&brm->rt_mem->descriptors[descriptor].buf_A);
	else
		offset = (uint32_t)
			ioread16be(&brm->rt_mem->descriptors[descriptor].buf_B);


	wc = (uint32_t) BRM_DATAWORDCOUNT(BRM_GET_WCMC(brm->brm_mem[offset]));

	return wc;
}


/**
 * @brief redirects PUS packets into buffers by source ID
 * @param brm a struct brm_config
 * @param rx_bytes number of bytes in the internal buffer
 * @retval  0 : success
 * @retval -1 : error
 */

static int32_t as250_split_packet_stream(struct brm_config *brm,
					 uint32_t rx_bytes)
{
	int32_t ret;
	uint32_t i = 0;
	uint32_t bytes;

	struct packet_tracker *p_pkt = NULL;

	while (i < rx_bytes) {

		bytes = (uint32_t) ptrack_get_pkt_size(&brm->buf.adb_ddb[i]);

		if (bytes > (AS250_TRANSFER_FRAME_SIZE-i)) {
			/* discard remaining buffer and return */
			errno = E_BRM_INVALID_PKT_SIZE;
			return -1;
		}

		switch (ptrack_get_next_pkt_src_id(&brm->buf.adb_ddb[i])) {

		case PUS_ID_GND:
			p_pkt = brm->buf.pkt_gnd;
			break;

		case PUS_ID_OBC:
			p_pkt = brm->buf.pkt_obc;
			break;

		default:
			/* discard remaining buffer and return */
			errno = E_BRM_INVALID_PKT_ID;
			return -1;
		}

		ret = ptrack_add_single_pkt(p_pkt, &brm->buf.adb_ddb[i],
					    (rx_bytes-i));

		if (ret < 0) {
			brmstat.bytes_dropped += rx_bytes;
			ptrack_reset(p_pkt);	/* reset buffers */
			return -1;	/* errno set by call to ptrack_add */
		}

		i += (uint32_t) ret;
	}

	return 0;
}


/**
 * @brief handles bus controller access to the AS250 DTD subaddress
 * @param brm a struct brm_config
 * @retval  0 : success
 * @retval -1 : error
 * @note ping-pong mode must already be disabled when this function is called
 * @note There are unnecessary copy operations when collecting the (packet)
 *       data. Unfortunately, this is not easily avoided, so we won't try and
 *       implement a zero-copy approach given the low data rate
 */

static int32_t as250_handle_dtd_irq(struct brm_config *brm)
{

	uint32_t i;
	uint32_t bytes;
	uint32_t rx_bytes_left;
	uint32_t offset;
	uint32_t rx_bytes = 0;

	struct transfer_descriptor dtd;


	offset = brm_get_offset_pp_rx_prev(brm, AS250_SA_RX_DTD);
	memcpy(&dtd, &brm->brm_mem[offset], AS250_TRANSFER_DESCRIPTOR_SIZE);

	/*  MIL-228/T; CIS-SW-ICD-035 & CIS-SW-ICD-036 protocol reset */
	if (unlikely(dtd.reset)) {
		brm->as250.dtd_blk_cnt = dtd.block_counter;

		dtd.qos_err   = 0x0;	/* MIL-228/T */
		dtd.reset     = 0x1;	/* MIL-228/T */

		goto send_dtc;
	}

	/*  variation from MIL-231/T in CIS-SW-ICD-033:
	 *  accept all block counters not identical to the last one as new
	 */
	if (likely(dtd.block_counter == brm->as250.dtd_blk_cnt))
		return 0;

	brm->as250.dtd_blk_cnt = dtd.block_counter;

	dtd.qos_err = 0x0;	/* default: no error detected */

	rx_bytes_left = dtd.transfer_size;

	for (i = AS250_SA_RX_DDB_01; i <= AS250_SA_RX_DDB_16; i++) {

		if (!rx_bytes_left)
			break;

		offset = brm_get_offset_pp_rx_prev(brm, i);

		if (rx_bytes_left > 64)
			bytes = 64;
		else
			bytes = rx_bytes_left;

		memcpy(&brm->buf.adb_ddb[rx_bytes],
		       &brm->brm_mem[offset], bytes);

		rx_bytes       += bytes;
		rx_bytes_left  -= bytes;
	}


send_dtc:	/* common exit */

	offset = brm_get_offset_pp_tx_next(brm, AS250_SA_TX_DTC);

	memcpy(&brm->brm_mem[offset], &dtd, AS250_TRANSFER_DESCRIPTOR_SIZE);

	brmstat.rx_bytes += rx_bytes;

	if (as250_split_packet_stream(brm, rx_bytes) < 0)
		return -1; /* errno set in call */

	return 0;
}


/**
 * @brief writes a protocol reset ATR
 * @param brm a struct brm_config
 * @retval 0 : success
 * @retval 1 : error occured
 *
 */

static uint32_t as250_reset_protocol(struct brm_config *brm)
{
	uint32_t offset;
	uint16_t pp_enabled = 0;


	if (brm_pp_enabled(brm)) {
		pp_enabled = 1;

		brm_disable_pp(brm);
#if (__sparc__) /* This is impossible to unit-test as long as we have static
		 * functions and must include the c file in the unit test file.
		 * We can also not just remove it, because it is mandated by
		 * hardware manual, see Core1553BRM v4.0 Handbook, p.61
		 */

		if (brm_pp_enabled(brm)) {
			errno = EFAULT;
			return 1;
		}
#endif
	}


	brm->as250.atr.word0     = 0x0000;
	brm->as250.atr.word1     = 0x4000; /* reset = 1 CIS-SW-ICD-036 */

	brm->as250.atd_blk_cnt = 0x1;

	offset = brm_get_offset_pp_tx_next(brm, AS250_SA_TX_ATR);
	memcpy(&brm->brm_mem[offset], &brm->as250.atr,
	       AS250_TRANSFER_DESCRIPTOR_SIZE);

	offset = brm_get_offset_pp_tx_curr(brm, AS250_SA_TX_ATR);
	memcpy(&brm->brm_mem[offset], &brm->as250.atr,
	       AS250_TRANSFER_DESCRIPTOR_SIZE);

	brm->as250.atr.word0     = 0x0;
	brm->as250.atr.word1     = 0x0;

	if (pp_enabled) {
		brm_enable_pp(brm);

#if (__sparc__) /* see above */
		if (!brm_pp_enabled(brm)) {
			errno = EFAULT;
			return 1;
		}
#endif
	}

	return 0;
}


/**
 * @brief creates a AS250 transfer frame
 * @param brm a struct brm_config
 * @param atr a struct transfer_descriptor
 * @retval -1 : error
 * @retval  0 : success
 */

static int32_t as250_create_transfer_frame(struct brm_config *brm,
					   struct transfer_descriptor *atr)
{
	int32_t ret;
	int32_t bytes;
	int32_t tx_bytes_left;

	struct cpus_buffers *p_buf;


	p_buf = brm->buf.atr_cpus;

	tx_bytes_left     = AS250_TRANSFER_FRAME_SIZE;
	atr->transfer_size = 0;

	while ((bytes = cpus_next_valid_pkt_size(p_buf))) {

		if (unlikely(bytes < 0)) {
			/* no point in counting dropped bytes here, as the
			 * buffer is not valid and it has to be reset as a
			 * whole
			 */
			cpus_reset(p_buf);
			return -1; /* errno set in call */
		}

		if (bytes > tx_bytes_left)
			break;

		ret = cpus_pop_packet(p_buf,
				      &brm->buf.adb_ddb[atr->transfer_size]);

		if (unlikely(ret < 1)) {
			cpus_reset(p_buf);
			return -1; /* errno set in call */
		}

		atr->transfer_size += ret;
		tx_bytes_left     -= ret;
	}

	return 0;
}


/**
 * @brief prepares the AS250 acquisition data buffer
 * @param brm a struct brm_config
 * @param atr a struct transfer_descriptor
 */

static void as250_prepare_adb(struct brm_config *brm,
			      const struct transfer_descriptor *atr)
{
	int32_t bytes;
	int32_t tx_bytes_left;
	int32_t tx_bytes;

	uint32_t offset;
	uint32_t sa = AS250_SA_TX_ADB_01;


	tx_bytes_left = atr->transfer_size;
	tx_bytes = 0;

	/* CIS-SW-ICD-042
	 * if data bytes transferred are not a multiple of two,
	 * there is one padding byte, clear in ADB
	 */
	if (tx_bytes_left & 0x1) {
		tx_bytes_left += 1;
		brm->buf.adb_ddb[tx_bytes_left] = 0x0;
	}

	do {
		if (!tx_bytes_left)
			break;

		offset = brm_get_offset_pp_tx_next(brm, sa);

		if (tx_bytes_left >= BRM_NUM_DATABYTES)
			bytes = BRM_NUM_DATABYTES;
		else
			bytes = tx_bytes_left;

		memcpy(&brm->brm_mem[offset],
		       &brm->buf.adb_ddb[tx_bytes], (size_t) bytes);

		tx_bytes      += bytes;
		tx_bytes_left -= bytes;

	} while (sa++ < AS250_SA_TX_ADB_16);
}

/**
 * @brief handles bus controller access to the AS250 ATR subaddress
 * @param brm a struct brm_config
 * @retval -1 : error
 * @retval  0 : success
 * @note ping-pong mode must already be disabled when this function is called
 * @note There are unnecessary copy operations when writing the data. This is
 *       relatively easy to fix, we just need a partial read-out mechanism
 *       for the buffers. It is however not an issue given the data rates.
 */

static int32_t as250_handle_atr_irq(struct brm_config *brm)
{
	int32_t ret;

	uint32_t offset;


	/* Set the last buffer accessed to the last valid ATR, otherwise the
	 * BC will consider the toggling block counter as an error because of
	 * the PP buffers.
	 */
	offset = brm_get_offset_pp_tx_next(brm, AS250_SA_TX_ATR);
	memcpy(&brm->brm_mem[offset], &brm->as250.atr,
	       AS250_TRANSFER_DESCRIPTOR_SIZE);

	ret = as250_create_transfer_frame(brm, &brm->as250.atr);
	if (ret)
		return -1; /* errno set in call */

	if (!brm->as250.atr.transfer_size)
		return 0;

	/**
	 * CIS-SW-ICD-030
	 * CIS-SW-ICD-031
	 *  MIL-224/T:     mode = 0 is default
	 *  MIL-223/T: set mode = 0 for SA 11-26 transfers
	 *		    qos = 1 verified data length
	 *
	 * note: MIL-251 / T states (n+1) modulo 256, but CASA wants wrap to 1
	 */

	/* AS250_USE_COUNTER_AS_SPECIFIED_NOT_AS_INTENDED_BY_CASA */
	brm->as250.atr.word1 = (0x8B00 | brm->as250.atd_blk_cnt++);

	as250_prepare_adb(brm, &brm->as250.atr);

	/* set the ATR */
	offset = brm_get_offset_pp_tx_next(brm, AS250_SA_TX_ATR);
	memcpy(&brm->brm_mem[offset], &brm->as250.atr,
	       AS250_TRANSFER_DESCRIPTOR_SIZE);

	brmstat.tx_bytes += brm->as250.atr.transfer_size;

	return 0;
}


/**
 * @brief handles bus controller access to the wrap-around subaddress
 * @param brm a struct brm_config
 */

static void as250_handle_wrap_irq(struct brm_config *brm)
{
	uint32_t bytes;
	uint32_t offset_rx;
	uint32_t offset_tx;

	offset_rx = brm_get_offset_pp_rx_prev(brm, AS250_SA_RX_DWRAP);
	offset_tx = brm_get_offset_pp_tx_next(brm, AS250_SA_TX_DWRAP);

	bytes = brm_get_wordcount(brm, AS250_SA_RX_DWRAP) * sizeof(uint16_t);

	memcpy(&brm->brm_mem[offset_tx], &brm->brm_mem[offset_rx], bytes);
}


/**
 * @brief handles direct time message subaddress access
 * @param brm a struct brm_config
 */

static void as250_handle_time_message_irq(struct brm_config *brm)
{
	uint32_t offset_rx;
	struct time_packet time;

	offset_rx = brm_get_offset_pp_rx_prev(brm, AS250_SA_RX_TIME);

	memcpy(&time, &brm->brm_mem[offset_rx], AS250_TIME_PACKET_SIZE);

	if (brm->set_time != NULL)
		brm->set_time(time.coarse_time, time.fine_time,
			      brm->set_time_userdata);
}


/**
 * @brief sets the first word in the RT_Health subaddress
 * @param brm a struct brm_config
 * @param mask a bit mask to set
 */

static void as250_rt_health_set(struct brm_config *brm, uint16_t mask)
{
	uint32_t offset;

	/* must always set both, they are not updated on access */
	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_A);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] = mask;

	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_B);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] = mask;
}


/**
 * @brief sets rt_health bits by mask
 * @param brm a struct brm_config
 * @param mask a mask with bits to toggle set HIGH
 */

static void as250_rt_health_set_bits(struct brm_config *brm, uint16_t mask)
{
	uint32_t offset;

	/* must always set both, they are not updated on access */
	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_A);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] |= mask;

	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_B);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] |= mask;
}


/**
 * @brief unsets rt_health bits by mask
 * @param brm a struct brm_config
 * @param mask a mask with bits to clear set HIGH
 */

static void as250_rt_health_unset_bits(struct brm_config *brm, uint16_t mask)
{
	uint32_t offset;

	/* must always set both, they are not updated on access */
	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_A);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] &= ~mask;

	offset  = (uint32_t)
		ioread16be(&brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].buf_B);
	offset += BRM_HEADER_WORDS;
	brm->brm_mem[offset] &= ~mask;
}


/**
 * @brief handles terminal configuration commands as per MIL-345/T to MIL-350/T
 * @param brm a struct brm_config
 */

static void as250_handle_termcfg(struct brm_config *brm)
{
	uint16_t rt_health_mask;
	uint32_t offset;

	/* no need to check for wordcount here */
	offset  = brm_get_offset_pp_rx_prev(brm, AS250_SA_RX_TERMCFG);
	rt_health_mask = brm->brm_mem[offset];

	as250_rt_health_unset_bits(brm, rt_health_mask);
}


/**
 * @brief RX subaddress IRQ handler for AS250 & ping-pong mode
 * @param brm a struct brm_config
 * @param descriptor a subaddress/mode code descriptor
 */

static void as250_handle_rx_irq(struct brm_config *brm, uint32_t descriptor)
{
	uint16_t flags;

	switch (BRM_CALC_SA_RX_DESC(descriptor)) {

	case AS250_SA_RX_TERMCFG:
		as250_handle_termcfg(brm);
		break;

	/* This is another case of the driver acting too fast. It appears that
	 * in cases where the DTD SA interrupt is raised by the 1553 core and
	 * the driver deactivates ping-pong mode, a race condition may occur,
	 * where the buffer A/B bit may have been toggled by the core, but the
	 * corresponding buffer is not, so the data may end up in the wrong
	 * buffer and is consequently not read by the driver.
	 * T̶h̶i̶s̶ ̶i̶s̶ ̶f̶i̶x̶e̶d̶ ̶b̶y̶ ̶c̶h̶a̶n̶g̶i̶n̶g̶ ̶t̶h̶e̶ ̶i̶n̶t̶e̶r̶r̶u̶p̶t̶i̶n̶g̶ ̶s̶u̶b̶a̶d̶d̶r̶e̶s̶s̶ ̶t̶o̶ ̶D̶T̶D̶
	 * s̶u̶b̶a̶d̶d̶r̶e̶s̶s̶ ̶1̶6̶.̶
	 * ^- This was a misinterpretation of the minor frame SA access scheme
	 *    and has since been reverted to the previous (correct) trigger on
	 *    DTD SA access.
	 *
	 * Note that this behaviour is only observed with the CHEOPS spacecraft
	 * platform simulator and is p̶o̶s̶s̶i̶b̶l̶y̶ definitely related to the
	 * asynchronous messaging scheme used, as the ADB/DDB SAs are only
	 * accessed conditionally if the transfer descriptors change and if a
	 * transfer fits into the data field of a the first subaddress, only
	 * that particular subaddress is read. This pattern does not continue,
	 * i.e. all transfers > 64 bytes will be executed as a read of all 16
	 * DDB subaddresses. This is most likely the reason why the race
	 * condition is usually only observed with packets <= 64 bytes.
	 *
	 * An optional delay has been added for such transfers. For performance
	 * reasons, this may be left out once we have final confirmation
	 * whether the actual spacecraft link will use asynchronous messaging
	 * the same way the platform simulator does.
	 *
	 * milFrameDelay is an external global and its information is defined 
	 * in the data pool. The global is constantly updated by the cyclical 
	 * task. 
	 */

	case AS250_SA_RX_DTD:
		if (as250_handle_dtd_irq(brm) < 0)
			brm_handle_error(brm, LOW);

		break;

	case AS250_SA_RX_TIME:
		/* CIS-SW-ICD-037 */
		flags  = ioread16be(&brm->rt_mem->sa_rx[AS250_SA_RX_TIME].CW)
			 & BRM_RT_CW_FLAGS_BRD;

		if (!flags)
			as250_handle_time_message_irq(brm);

		break;

	case AS250_SA_RX_DWRAP:
		as250_handle_wrap_irq(brm);
		break;

	default:
		break;
	}
}


/**
 * @brief TX subaddress IRQ handler for AS250 & ping-pong mode
 * @param brm a struct brm_config
 * @param descriptor a subaddress/mode code descriptor
 */

static void as250_handle_tx_irq(__attribute__((unused)) struct brm_config *brm,
				uint32_t descriptor)
{
	switch (BRM_CALC_SA_TX_DESC(descriptor)) {

	case AS250_SA_TX_ATR:
		break;

	default:
		break;
	}
}

/**
 * @brief mode-code RX subaddress IRQ handler for AS250 & ping-pong mode#
 * @param brm a struct brm_config
 * @param descriptor a subaddress/mode code descriptor
 */

static void as250_handle_mc_rx_irq(struct brm_config *brm, uint32_t descriptor)
{
	uint16_t frame_num;

	switch (BRM_CALC_MC_RX_DESC(descriptor)) {

	case AS250_MC_RESET_RT:
		if (brm->rt_reset)
			brm->rt_reset(brm->rt_reset_userdata);

		break;

	case AS250_MC_SYNC_W_DWORD:
		frame_num = brm->brm_mem[brm_get_offset_pp_mc_rx_prev(brm,
						 AS250_MC_SYNC_W_DWORD)];
		if (brm->rt_sync_with_dword)
			brm->rt_sync_with_dword(frame_num,
					brm->rt_sync_with_dword_userdata);
	/* The driver is too fast for the hardware, at least on the GR712RC
	 * development board. We would overwrite the ADB SA buffers
	 * before the hardware reads it, because the the ATR SA irq is tripped
	 * and its handler executed before the 1553 core can ping-pong even
	 * the first buffer.
	 * Usually, we would act after the core accesses the last ADB SA, but
	 * since the PS polls the ADB SAs only on condition a ATR has been set
	 * (which the real OBC probably won't do), we can't do that and the
	 * sync-with-dataword modecode is the only other option.
	 * Note: this occurs only with high transfer rates, i.e. when setting an
	 * ATR every minor frame
	 */

		if (as250_handle_atr_irq(brm) < 0)
			brm_handle_error(brm, LOW);

		break;

	default:
		break;
	}
}


/**
 * @brief subaddress IRQ handler for ping-pong mode
 * @param brm a struct brm_config
 * @param descriptor a subaddress/mode code descriptor
 * @retval -1 : error
 * @retval  0 : success
 * @note uses GNU extension: range expression in switch-statement
 */

extern unsigned int milFrameDelay;

__extension__
static int32_t brm_irq_subad_pp(struct brm_config *brm, uint32_t descriptor)
{
	uint16_t cw;
	struct brm_descriptor_table *brm_descriptor;


	brm_descriptor = &brm->rt_mem->descriptors[descriptor];

	cw = ioread16be(&brm_descriptor->CW);

	if (!(cw & BRM_RT_CW_FLAGS_BAC)) {
		errno = E_BRM_CW_BAC_FLAG;
		return -1;
	}

	/* see as250_handle_rx_irq() */
	/* 4000 amounts to about 160 µs */
	/* yes, this must be volatile */
	{
	  volatile uint32_t i = 0;
	  volatile uint32_t EXTRA_DELAY_BEFORE_PP_DISABLE;

	  EXTRA_DELAY_BEFORE_PP_DISABLE = milFrameDelay;

	  while (i++ < EXTRA_DELAY_BEFORE_PP_DISABLE)
	    __asm__ __volatile__("nop");
	}


	brm_disable_pp(brm);

#if (__sparc__) /* see as250_reset_protocol() for rationale */
	if (brm_pp_enabled(brm)) {
		errno = EFAULT;
		return -1;
	}
#endif

	/*  see Core1553BRM v4.0 Handbook, pp.55 for magic numbers */
	switch (descriptor) {
	case  1 ...  32:
		as250_handle_rx_irq(brm, descriptor);
		break;
	case 33 ...  64:
		as250_handle_tx_irq(brm, descriptor);
		break;
	case 65 ...  96:
		as250_handle_mc_rx_irq(brm, descriptor);
		break;
	case 97 ... 128:
		/* unused (?) */
		/* as250_handle_mc_tx_irq(brm); */
		break;
	default:
		break;
	}

	cw &= ~BRM_RT_CW_FLAGS_BAC;

	iowrite16be(cw, &brm_descriptor->CW);

	brm_enable_pp(brm);

#if (__sparc__) /* see above */
	if (!brm_pp_enabled(brm)) {
		errno = EFAULT;
		return -1;
	}
#endif

	return 0;
}


/**
 * @brief checks and handles error flags set in a irq log entry
 * @param brm a struct brm_config
 * @param BRM_IRQ error flags to check
 */

__attribute__((noinline))
static void brm_handle_log_error(struct brm_config *brm, uint16_t BRM_IRQ)
{
	if (BRM_IRQ & BRM_IRQ_RT_ILLCMD) {
		errno = E_BRM_IRQ_RT_ILLCMD;
		brm_handle_error(brm, MEDIUM);
	}

	if (BRM_IRQ & BRM_IRQ_ILLOP) {
		errno = E_BRM_IRQ_ILLOP;
		brm_handle_error(brm, MEDIUM);
	}

	if (BRM_IRQ & BRM_IRQ_MERR) {
		errno = E_BRM_IRQ_MERR;
		brm_handle_error(brm, MEDIUM);
	}
}


/**
 * @brief checks and handles error flags set in the hw irq register
 * @param brm a struct brm_config
 * @param BRM_IRQ error flags to check
 */

__attribute__((noinline))
static void brm_handle_irq_error(struct brm_config *brm, uint16_t BRM_IRQ)
{
	if (BRM_IRQ & BRM_IRQ_DMAF) {
		errno = E_BRM_IRQ_DMAF;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_WRAPF) {
		errno = E_BRM_IRQ_WRAPF;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_TAPF) {
		errno = E_BRM_IRQ_TAPF;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_BITF) {
		errno = E_BRM_IRQ_BITF;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_MERR) {
		errno = E_BRM_IRQ_MERR;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_IXEQ0) {
		errno = E_BRM_IRQ_IXEQ0;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_RT_ILLCMD) {
		errno = E_BRM_IRQ_RT_ILLCMD;
		brm_handle_error(brm, HIGH);
	}

	if (BRM_IRQ & BRM_IRQ_ILLOP) {
		errno = E_BRM_IRQ_ILLOP;
		brm_handle_error(brm, HIGH);
	}
}


/**
 * @brief the main IRQ handler
 * @param userdata a pointer to arbitrary data
 * @retval 0 : always
 */

__attribute__((unused))
static int32_t brm_irq_handler(void *userdata)
{
	uint16_t BRM_IRQ;

	uint32_t descriptor;
	uint32_t flags;

	int32_t ret;

	static uint32_t pos;

	struct brm_config *brm;
	struct brm_irq_log_entry *log;


	brm = (struct brm_config *) userdata;
	log = brm->rt_mem->irq_log;

	while ((BRM_IRQ = ioread16be(&log[pos].IIW))
			!= BRM_IIW_PATTERN_INVALID) {

		iowrite16be(BRM_IIW_PATTERN_INVALID, &log[pos].IIW);

		flags = BRM_IRQ_MERR | BRM_IRQ_RT_ILLCMD | BRM_IRQ_ILLOP;

		if (BRM_IRQ & flags)
			brm_handle_log_error(brm, BRM_IRQ);

		if (BRM_IRQ & BRM_IRQ_SUBAD) {
			/* offset to address */
			descriptor = ioread16be(&log[pos].IAW) >> 2;

			ret = brm_irq_subad_pp(brm, descriptor);
			if (ret < 0)
				brm_handle_error(brm, MEDIUM);

		}

		BRM_LOGLIST_INCREMENT_READPOS(pos, BRM_IRQ_LOG_ENTRY_PAGES);
	}


	/*  handle hardware IRQs (processing cannot be deferred,
	 *  they are not logged)
	 */

	BRM_IRQ = ioread16be(&brm->regs->ipend);

	flags = BRM_IRQ_DMAF | BRM_IRQ_WRAPF | BRM_IRQ_TAPF | BRM_IRQ_BITF
		| BRM_IRQ_MERR | BRM_IRQ_IXEQ0 | BRM_IRQ_RT_ILLCMD;

	/* this will be called often, so do one quick check and detailed
	 * ones only if needed
	 */

	if (BRM_IRQ & flags)
		brm_handle_irq_error(brm, BRM_IRQ);

	return 0;
}


/**
 * @brief sets the function to call when a time message arrives
 * @param brm a struct brm_config
 * @param callback a callback function
 * @param userdata a pointer to arbitrary data supplied to the callback function
 * @retval  0 : success
 * @retval -1 : error
 */

int32_t as250_set_time_callback(struct brm_config *brm,
				void (*callback)(uint32_t coarse_time,
						 uint32_t fine_time,
						 void *userdata),
				void *userdata)
{
	if (!callback) {
		errno = EINVAL;
		return -1;
	}

	brm->set_time = callback;
	brm->set_time_userdata = userdata;

	return 0;
}

/**
 * @brief sets the function to call when a Reset RT mode code arrives
 * @param brm a struct brm_config
 * @param callback a callback function
 * @param userdata a pointer to arbitrary data supplied to the callback function
 * @retval  0 : success
 * @retval -1 : error
 */

int32_t as250_set_rt_reset_callback(struct brm_config *brm,
				    void (*callback)(void *userdata),
				    void *userdata)
{
	if (!callback) {
		errno = EINVAL;
		return -1;
	}

	brm->rt_reset = callback;
	brm->rt_reset_userdata = userdata;

	return 0;
}

/**
 * @brief sets the function to call when a Sync With Dataword mode code arrives
 * @param brm a struct brm_config
 * @param callback a callback function
 * @param userdata a pointer to arbitrary data supplied to the callback function
 * @retval  0 : success
 * @retval -1 : error
 */

int32_t as250_set_sync_with_dword_callback(struct brm_config *brm,
					   void (*callback)(uint16_t frame_num,
							    void *userdata),
					   void *userdata)
{
	if (callback == NULL) {
		errno = EINVAL;
		return -1;
	}

	brm->rt_sync_with_dword = callback;
	brm->rt_sync_with_dword_userdata = userdata;

	return 0;

}


/**
 * @brief sets up the ping pong buffers addresses and interrupt flags as
 *	  needed for AS250
 * @param brm a struct brm_config
 */

static void brm_rt_configure_pp_buffers(struct brm_config *brm)
{
	uint32_t i, j;
	uint16_t flags;

	struct brm_message_circular_buffer *msg_buf[4];
	struct brm_descriptor_table        *desc[4];


	msg_buf[0] = brm->rt_mem->sa_msgs_rx;
	msg_buf[1] = brm->rt_mem->sa_msgs_tx;
	msg_buf[2] = brm->rt_mem->mc_msgs_rx;
	msg_buf[3] = brm->rt_mem->mc_msgs_tx;

	desc[0] = brm->rt_mem->sa_rx;
	desc[1] = brm->rt_mem->sa_tx;
	desc[2] = brm->rt_mem->mc_rx;
	desc[3] = brm->rt_mem->mc_tx;


	for (i = 0; i < 4; i++) {
		for (j = 0; j < BRM_NUM_SUBADDR; j++) {
			desc[i][j].buf_A = BASE_ADDR(msg_buf[i][j].page[0]);

			desc[i][j].buf_B = BASE_ADDR(msg_buf[i][j].page[1]);

			desc[i][j].bcast = BASE_ADDR(msg_buf[i][j].page[0]);
		}
	}

	flags = BRM_RT_CW_FLAGS_IWA | BRM_RT_CW_FLAGS_AB;

	iowrite16be(flags, &brm->rt_mem->sa_rx[AS250_SA_RX_DTD].CW);
	iowrite16be(flags, &brm->rt_mem->sa_rx[AS250_SA_RX_TIME].CW);
	iowrite16be(flags, &brm->rt_mem->sa_rx[AS250_SA_RX_TERMCFG].CW);
	iowrite16be(flags, &brm->rt_mem->sa_rx[AS250_SA_RX_DWRAP].CW);

	iowrite16be(flags, &brm->rt_mem->mc_rx[AS250_MC_SYNC_W_DWORD].CW);

	flags = BRM_RT_CW_FLAGS_AB;

	iowrite16be(flags, &brm->rt_mem->sa_tx[AS250_SA_TX_HLTHMON].CW);

	flags = BRM_RT_CW_FLAGS_IWA | BRM_RT_CW_FLAGS_IBRD | BRM_RT_CW_FLAGS_AB;
	iowrite16be(flags, &brm->rt_mem->mc_rx[AS250_MC_RESET_RT].CW);


	flags  = BRM_CCR_BAEN | BRM_CCR_BBEN | BRM_CCR_BCEN
		 | BRM_CCR_INEN | BRM_CCR_PPEN;

	iowrite32be(flags, &brm->regs->ctrl);
}


/**
 * @brief sets desynchronized flag in rt_health
 * @param brm a struct brm_config
 * @note to be called by external timing management
 */

void as250_desynchronized(struct brm_config *brm)
{
	as250_rt_health_set_bits(brm, AS250_RT_NOT_SYNCHRIONIZED);
}


/**
 * @brief unsets desynchronized flag in rt_health
 * @param brm a struct brm_config
 * @note to be called by external timing management
 */

void as250_synchronized(struct brm_config *brm)
{
	as250_rt_health_unset_bits(brm, AS250_RT_NOT_SYNCHRIONIZED);
}


/**
 * @brief configures the remote terminal address and parity bit
 * @param brm a struct brm_config
 * @param rt_address the remote terminal address to set
 */

static void brm_rt_config_addr(struct brm_config *brm, uint8_t rt_address)
{
	uint32_t flags;

	flags = BRM_CCR_RTA(rt_address)
	      | (rt_parity_bit(rt_address) << BRM_CCR_RTPTY_REG)
	      | BRM_OSR_MSEL_RT;

	iowrite32be(flags, &brm->regs->oper);
}


/**
 * @brief configures the remote terminal's irq mask
 * @param brm a struct brm_config
 */

static void brm_rt_config_irqmask(struct brm_config *brm)
{
	uint32_t flags;

	flags = BRM_IRQ_RT_ILLCMD | BRM_IRQ_SUBAD | BRM_IRQ_TAPF
		| BRM_IRQ_DMAF | BRM_IRQ_WRAPF | BRM_IRQ_MERR;

	iowrite32be(flags, &brm->regs->imask);
}


/**
 * @brief configures descriptor/command block pointer
 * @param brm a struct brm_config
 */

static void brm_setup_desc_pointer(struct brm_config *brm)
{
	/* point to base of memory area */
	iowrite32be(0x0, &brm->regs->dpoint);
}


/**
 * @brief configures irq log list pointer
 * @param brm a struct brm_config
 */

static void brm_setup_irq_pointer(struct brm_config *brm)
{
	uint32_t addr;

	/* see p78 Core1553BRM Manual */
	addr = BASE_ADDR(brm->rt_mem->irq_log[0]);

	iowrite32be(addr, &brm->regs->ipoint);
}


/**
 * @brief informs 1553 core about configured core frequency
 * @param brm a struct brm_config
 * @param core_freq the core frequency to set
 * @retval -1 : error
 * @retval  0 : success
 */

static uint32_t brm_setup_core_freq(struct brm_config *brm, uint32_t core_freq)
{
	switch (core_freq) {
	case BRM_FREQ_12MHZ:
	case BRM_FREQ_16MHZ:
	case BRM_FREQ_20MHZ:
	case BRM_FREQ_24MHZ:
		break;	/* all legal */
	default:
		errno = E_BRM_INVALID_COREFREQ;
		return 1;
	}

	iowrite32be(core_freq, &brm->regs->enhanced);

	return 0;
}


/**
 * @brief sets up hardware and message irq acknowledge
 * @param brm a struct brm_config
 * @note see GR712UM, p.162
 */

static void gr712_setup_1553_intack(struct brm_config *brm)
{
	uint32_t flags;

	flags = IRQ_ACK_MSG | IRQ_ACK_HW;

	iowrite32be(flags, &brm->regs->w_irqctrl);
}


/**
 * @brief sets up AHB page address register (i.e. 1553 base memory area)
 * @param brm a struct brm_config
 * @note see GR712UM, p.162
 */

static void gr712_setup_1553_ahbaddr(struct brm_config *brm)
{
	uint32_t addr;

	addr = (uint32_t) brm->brm_mem;

	iowrite32be(addr, &brm->regs->w_ahbaddr);
}

/**
 * @brief pulls up sub-system flag bit (toggles bit in 1553 core low),
 * @param brm a struct brm_config
 * @note see GR172UM, p.162
 */

static void gr712_set_ssysfn(struct brm_config *brm)
{
	uint32_t flags;

	flags  = ioread32be(&brm->regs->w_ctrl);
	flags |= 0x1;

	iowrite32be(flags, &brm->regs->w_ctrl);
}

#if (__sparc__)
/**
 * @brief configures core frequency on GR712RC eval board
 * @param brm a struct brm_config
 */

void gr712_set_core_freq(void)
{
	/* WONTFIX: magic numbers */
	uint32_t *gpreg = (uint32_t *) 0x80000600;
	uint32_t reg;

	/* set core frequency to 20 MHz */
	reg = ioread32be(gpreg) | (((2 << 16) | (0 << 14)) & (~(0 << 18)));

	iowrite32be(reg, gpreg);
}
#endif


/**
 * @brief illegalize all RT commands
 *
 * @param brm a struct brm_config
 */

static void brm_rt_illegalize_all(struct brm_config *brm)
{
	uint32_t i;
	uint32_t flags;


	for (i = 0; i < ARRAY_SIZE(brm->regs->rt_cmd_leg); i++) {
		BRM_LEG_CMD_ILLEGALIZE_ALL(flags);
		iowrite32be(flags, &brm->regs->rt_cmd_leg[i]);
	}
}


/**
 * @brief legalize a RT command
 *
 * @param brm a struct brm_config
 * @param reg the command register offset
 * @param cmd the command to legalize
 */

static void brm_rt_legalize_cmd(struct brm_config *brm,
				uint32_t reg, uint32_t cmd)
{
	uint32_t flags;

	flags =	ioread32be(&brm->regs->rt_cmd_leg[reg]);
	BRM_LEG_CMD_LEGALIZE(flags, cmd);
	iowrite32be(flags, &brm->regs->rt_cmd_leg[reg]);
}


/**
 * @brief legalize a rx subaddress
 *
 * @param brm a struct brm_config
 * @param sa the subaddress to legalize
 */

static void brm_rt_legalize_sa_rx(struct brm_config *brm, uint32_t sa)
{
	uint32_t reg = BRM_LEG_CMD_SA_RX__0_15;

	if (sa >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_SA_RX_16_31;
		sa -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, sa);
}


/**
 * @brief legalize a tx subaddress
 *
 * @param brm a struct brm_config
 * @param sa the subaddress to legalize
 */

static void brm_rt_legalize_sa_tx(struct brm_config *brm, uint32_t sa)
{
	uint32_t reg = BRM_LEG_CMD_SA_TX__0_15;

	if (sa >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_SA_TX_16_31;
		sa -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, sa);
}


/**
 * @brief legalize a broadcast rx subaddress
 *
 * @param brm a struct brm_config
 * @param sa the subaddress to legalize
 */

static void brm_rt_legalize_sa_bc_rx(struct brm_config *brm, uint32_t sa)
{
	uint32_t reg = BRM_LEG_CMD_SA_BC_RX__0_15;

	if (sa >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_SA_BC_RX_16_31;
		sa -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, sa);
}


/**
 * @brief legalize a broadcast tx subaddress
 *
 * @param brm a struct brm_config
 * @param sa the subaddress to legalize
 */

__attribute__((unused))
static void brm_rt_legalize_sa_bc_tx(struct brm_config *brm, uint32_t sa)
{
	uint32_t reg = BRM_LEG_CMD_SA_BC_TX__0_15;

	if (sa >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_SA_BC_TX_16_31;
		sa -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, sa);
}


/**
 * @brief legalize a rx mode code
 *
 * @param brm a struct brm_config
 * @param mc the mode code to legalize
 */

static void brm_rt_legalize_mc_rx(struct brm_config *brm, uint32_t mc)
{
	uint32_t reg = BRM_LEG_CMD_MC_RX__0_15;

	if (mc >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_MC_RX_16_31;
		mc -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, mc);
}


/**
 * @brief legalize a tx mode code
 *
 * @param brm a struct brm_config
 * @param mc the mode code to legalize
 */

__attribute__((unused))
static void brm_rt_legalize_mc_tx(struct brm_config *brm, uint32_t mc)
{
	uint32_t reg = BRM_LEG_CMD_MC_TX__0_15;

	if (mc >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_MC_TX_16_31;
		mc -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, mc);
}


/**
 * @brief legalize a broadcast rx mode code
 *
 * @param brm a struct brm_config
 * @param mc the mode code to legalize
 */

static void brm_rt_legalize_bc_mc_rx(struct brm_config *brm, uint32_t mc)
{
	uint32_t reg = BRM_LEG_CMD_MC_BC_RX__0_15;

	if (mc >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_MC_BC_RX_16_31;
		mc -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, mc);
}


/**
 * @brief legalize a broadcast tx mode code
 *
 * @param brm a struct brm_config
 * @param mc the mode code to legalize
 */

__attribute__((unused))
static void brm_rt_legalize_bc_mc_tx(struct brm_config *brm, uint32_t mc)
{
	uint32_t reg = BRM_LEG_CMD_MC_BC_TX__0_15;

	if (mc >= BMR_LEG_CMDS_PER_REGISTER) {
		reg = BRM_LEG_CMD_MC_BC_TX_16_31;
		mc -= BMR_LEG_CMDS_PER_REGISTER;
	}

	brm_rt_legalize_cmd(brm, reg, mc);
}


/**
 * @brief sets up legal RT commands
 * @param brm a struct brm_config
 * @note see Core1553BRM v4.0 Handbook, pp.83
 *
 * @note uses GNU extension: binary constants
 */

__extension__
static void brm_rt_legalize_as250_commands(struct brm_config *brm)
{
	brm_rt_illegalize_all(brm);

	/* subaddresses according to table 6.7-1 in DIV.SP.00030.T.ASTR */

	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_TERMCFG);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_01);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_02);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_03);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_04);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_05);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_06);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_07);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_08);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_09);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_10);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_11);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_12);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_13);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_14);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_15);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DDB_16);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DTD);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_TIME);
	brm_rt_legalize_sa_rx(brm, AS250_SA_RX_DWRAP);

	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_HLTHMON);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_01);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_02);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_03);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_04);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_05);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_06);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_07);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_08);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_09);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_10);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_11);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_12);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_13);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_14);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_15);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ADB_16);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_DTC);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_ATR);
	brm_rt_legalize_sa_tx(brm, AS250_SA_TX_DWRAP);

	brm_rt_legalize_sa_bc_rx(brm, AS250_SA_RX_TIME);

	/* mandatory mode codes according to table 6.3-1 in
	 * DIV.SP.00030.T.ASTR
	 */


	brm_rt_legalize_mc_rx(brm, BRM_MC_TRANSMIT_STATUS_WORD);
	brm_rt_legalize_mc_rx(brm, BRM_MC_TRANSMITTER_SHUTDOWN);
	brm_rt_legalize_mc_rx(brm, BRM_MC_OVERRIDE_TRANSMITTER_SHUTDOWN);
	brm_rt_legalize_mc_rx(brm, BRM_MC_RESET_REMOTE_TERMINAL);

	/* UVIE bus controller emulation has a broadcast address issue, this may
	 * be driver related
	 */
	brm_rt_legalize_mc_rx(brm, BRM_MC_SYNCHRONIZE_WITH_DATA_WORD);

	/* sync with dataword marks the start of a frame, see CIS-SW-ICD-032
	 * in CHEOPS-ECE-CSW-ICD-049
	 */
	brm_rt_legalize_bc_mc_rx(brm, BRM_MC_SYNCHRONIZE_WITH_DATA_WORD);
}


/**
 * @brief clears RT irq log
 * @param brm a struct brm_config
 */

static void brm_rt_clear_irq_log(struct brm_config *brm)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(brm->rt_mem->irq_log); i++) {
		iowrite16be(BRM_IIW_PATTERN_INVALID,
			    &brm->rt_mem->irq_log[i].IIW);
		iowrite16be(0x0, &brm->rt_mem->irq_log[i].IAW);
	}
}


/**
 * @brief configures the desired RT operating mode
 * @param brm a struct brm_config
 * @retval 1 : error
 * @retval 0 : success
 * @note  only ping-pong is enabled in this version of the driver
 */

static uint32_t brm_rt_configure_mode(struct brm_config *brm)
{

	switch (brm->mode) {
	case PING_PONG:
		brm_rt_configure_pp_buffers(brm);
		break;
	default:
		errno = EINVAL;
		return 1;
	}

	return 0;
}


/**
 * @brief soft-resets remote terminal
 * @param brm a struct brm_config
 */

static void brm_rt_softreset(struct brm_config *brm)
{
	uint32_t flags;

	flags  = ioread32be(&brm->regs->ctrl);
	flags |= BRM_CCR_SRST;

	iowrite32be(flags, &brm->regs->ctrl);
}

/**
 * @brief starts remote terminal execution
 * @param brm a struct brm_config
 */

static void brm_rt_start(struct brm_config *brm)
{
	uint32_t flags;

	flags  = ioread32be(&brm->regs->ctrl);
	flags |= BRM_CCR_STEX;

	iowrite32be(flags, &brm->regs->ctrl);
}


/**
 * @brief points rt memory to allocated brm memory block
 * @param brm a struct brm_config
 */

static void brm_rt_set_mem(struct brm_config *brm)
{
	brm->rt_mem = (void *) brm->brm_mem;
}


/**
 * @brief intialises remote terminal
 * @param brm a struct brm_config
 * @param rt_address the address of the remote terminal
 * @param core_freq the core frequency to set
 * @retval 1 : error
 * @retval 0 : success
 */

uint32_t brm_rt_init(struct brm_config *brm,
		     uint8_t rt_address, uint32_t core_freq)
{
	__attribute__((unused))
	uint32_t ret;


	if (brm->buf.adb_ddb == NULL) {
		errno = EINVAL;
		return 1;
	}

	if (brm->buf.atr_cpus == NULL) {
		errno = EINVAL;
		return 1;
	}

	brm_rt_set_mem(brm);

	as250_rt_health_set(brm, 0x0);

	brm_rt_softreset(brm);

	gr712_set_ssysfn(brm);

	brm_rt_config_addr(brm, rt_address);

	brm_rt_config_irqmask(brm);

	brm_setup_desc_pointer(brm);

	brm_setup_irq_pointer(brm);

	if (brm_setup_core_freq(brm, core_freq))
		return 1; /* errno set in call */

	gr712_setup_1553_intack(brm);

	gr712_setup_1553_ahbaddr(brm);

	brm_rt_legalize_as250_commands(brm);

	brm_rt_clear_irq_log(brm);

	if (brm_rt_configure_mode(brm))
		return 1; /* errno set in call */


	brm_rt_start(brm);

	as250_rt_health_set_bits(brm, AS250_INITIALIZATION_COMPLETE);

	/**
	 * CIS-SW-ICD-034
	 * issue a protocol reset immediately at startup
	 */

	ret = as250_reset_protocol(brm);
#if (__sparc__) /* return value of failed brm_pp_enabled, cannot be tested
		 * see as250_reset_protocol() for rationale
		 */
	if (ret)
		return ret;
#endif

	return 0;
}


/**
 * @brief set the 1kib working buffer
 * @param brm a struct brm_config
 * @param p_buf the buffer to assign
 * @retval -1 : error
 * @retval  0 : success
 */

int32_t brm_set_adb_ddb_buffer(struct brm_config *brm, uint8_t *p_buf)
{
	if (p_buf != NULL) {
		brm->buf.adb_ddb = p_buf;
	} else {
		errno = EINVAL;
		return -1;
	}

	return 0;
}

/**
 * @brief set the packet tracker for OBC packets
 * @param brm a struct brm_config
 * @param packet_tracker a struct packet_tracker
 * @retval -1 : error
 * @retval  0 : success
 */

int32_t brm_set_packet_tracker_obc(struct brm_config *brm,
				   struct packet_tracker *pkt_obc)
{
	if (pkt_obc != NULL) {
		brm->buf.pkt_obc = pkt_obc;
	} else {
		errno = EINVAL;
		return -1;
	}

	return 0;
}

/**
 * @brief sets the packet tracker for GND packets
 * @param brm a struct brm_config
 * @param packet_tracker a struct packet_tracker
 * @retval -1 : error
 * @retval  0 : success
 */

int32_t brm_set_packet_tracker_gnd(struct brm_config *brm,
				   struct packet_tracker *pkt_gnd)
{
	if (pkt_gnd != NULL) {
		brm->buf.pkt_gnd = pkt_gnd;
	} else {
		errno = EINVAL;
		return -1;
	}

	return 0;
}

/**
 * @brief sets the sliding window circular PUS packet buffer
 * @param brm a struct brm_config
 * @param p_cpus a strct cpus_buffers
 * @retval -1 : error
 * @retval  0 : success
 */

int32_t as250_set_atr_cpus_buffers(struct brm_config *brm,
				   struct cpus_buffers *p_cpus)
{
	if (p_cpus != NULL) {
		brm->buf.atr_cpus = p_cpus;
	} else {
		errno = EINVAL;
		return -1;
	}

	return 0;
}



/**
 * @brief enables Core1553BRM
 * @param brm a struct brm_config
 * @retval -1 : error
 * @retval  0 : success
 */

int32_t brm_1553_enable(struct brm_config *brm)
{
	struct sysobj *sobj;
#if (__sparc__)
	/**
	 * Basic Moron Protector (BMP)™
	 * (assume single device, ignore error case)
	 */
	static enum  {DISABLED, ENABLED} brm1553;


	if (brm1553)
		return -1;

	brm1553 = ENABLED;
#endif
	if (!brm->regs) {
		errno = EINVAL;
		return -1;
	}

	if (!brm->brm_mem) {
		errno = EINVAL;
		return -1;
	}


	if (((uint32_t) brm->brm_mem) & BRM_MEM_BLOCK_ALIGN)	{
		errno = E_BRM_MEM_ADDR_ALIGN;
		return -1;
	}

	/* explicitly set them null in case the user forgot
	 * to zero  brm_config
	 */
	brm->set_time           = NULL;
	brm->rt_reset           = NULL;
	brm->rt_sync_with_dword = NULL;




	sobj = sysobj_create();

	if (!sobj)
		return -1;

	sobj->sattr = brm_attributes;

	sysobj_add(sobj, NULL, driver_set, "1553");


#if (__sparc__)
	irq_dispatch_enable();

	irl1_register_callback(GR712_IRL1_B1553BRM, PRIORITY_NOW,
			       &brm_irq_handler, (void *) brm);
#endif
	return 0;
}
