/**
 * @file   event_report.c
 * @ingroup event_report
 * @author Roland Ottensamer (roland.ottensamer@univie.ac.at),
 *          Armin Luntzer (roland.ottensamer@univie.ac.at)
 * @date   2015
 *
 * @copyright GPLv2
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * @defgroup event_report Central event report transation interface to IASW
 *
 * @brief IBSW error translation to IASW events
 *
 *
 * ## Overview
 *
 * Since it is required that a horizontal hierarchial topology with partially
 * reverse dependencies is forced on what is clearly vertical partitioning in
 * the supplied abstraction levels of hardware and software functionality of
 * the IBSW component, it is sensible to at least minimise the adaption points.
 *
 * This component provides a central translation interface and submission of
 * event reports generated by the IBSW to the IASW event reporting interface.
 *
 *
 * ## Mode of Operation
 *
 * This report arguments passed to this interface are (for obvious reasons)
 * similar to those used in the IASW. Since the IASW will crash if a
 * combination of error id and severity is not defined, the arguments are
 * filtered, so only legal combinations are passed to the IASW interface.
 *
 * If IASW definitions are changed, these filters must be updated accordingly.
 *
 * Given that the IASW errors do not actually contain useful information by
 * themselves, relevant error information (defined as errno-extensions) are
 * passed via the event data payload of an error.
 *
 *
 * ## Error Handling
 *
 * None
 *
 *
 * ## Notes
 *
 * From a design perspective, the IBSW should rather maintain its own
 * independent error message buffer, which is queried by the IASW as desired.
 * Instead, any error reported by the IBSW must invoke IASW functionality even
 * from interrupt contexts wich makes such calls unpredictable and effectively
 * destabilises the IBSW, as all IASW defects regarding event report and/or
 * packet manangement immediately-back propagate or even halt the operation of
 * the IBSW.
 *
 */


#include <event_report.h>


#if (!__sparc__)

void CrIaEvtRaise(__attribute__((unused)) uint8_t subtype,
		__attribute__((unused)) uint16_t id,
		__attribute__((unused)) uint16_t *data,
		__attribute__((unused)) uint32_t data_size)
{
}

#endif


/**
 * @brief translate normal level errors
 * @param errno an error number
 *
 * @return the IASW level identifier of the error class
 */

static uint16_t event_report_translate_class_norm(enum error_class c)
{
	uint16_t id = 0;

	switch (c) {
	case INIT:
		id = CRIA_SERV5_EVT_INIT_SUCC;
		break;
	case NOTIFY:
		break;
	case FBF:
		break;
	case EDAC:
		break;
	case CORE1553BRM:
		break;
	case GRSPW2:
		break;
	case ERRLOG:
		break;
	case SYNC:
		break;
	}

	return id;
}


/**
 * @brief translate low level errors
 * @param c the error class
 *
 * @return the IASW level identifier of the error class
 */

static uint16_t event_report_translate_class_low(enum error_class c)
{
	uint16_t id = 0;

	switch (c) {
	case INIT:
		break;
	case NOTIFY:
		break;
	case FBF:
		break;
	case EDAC:
		id = CRIA_SERV5_EVT_SBIT_ERR;
		break;
	case CORE1553BRM:
		id = CRIA_SERV5_EVT_1553_ERR_L;
		break;
	case GRSPW2:
		id = CRIA_SERV5_EVT_SPW_ERR_L;
		break;
	case ERRLOG:
		break;
	case SYNC:
		break;
	}

	return id;
}


/**
 * @brief translate medium level errors
 * @param c the error class
 *
 * @return the IASW level identifier of the error class
 */

static uint16_t event_report_translate_class_med(enum error_class c)
{
	uint16_t id = 0;

	switch (c) {
	case INIT:
		break;
	case NOTIFY:
		break;
	case FBF:
		break;
	case EDAC:
		id = CRIA_SERV5_EVT_DBIT_ERR;
		break;
	case CORE1553BRM:
		id = CRIA_SERV5_EVT_1553_ERR_M;
		break;
	case GRSPW2:
		id = CRIA_SERV5_EVT_SPW_ERR_M;
		break;
	case ERRLOG:
		break;
	case SYNC:
		id = CRIA_SERV5_EVT_SYNC_LOSS;
		break;
	}

	return id;
}


/**
 * @brief translate high level errors
 * @param c the error class
 *
 * @return the IASW level identifier of the error class
 */

static uint16_t event_report_translate_class_high(enum error_class c)
{
	uint16_t id = 0;

	switch (c) {
	case INIT:
		id = CRIA_SERV5_EVT_INIT_FAIL;
		break;
	case NOTIFY:
		break;
	case FBF:
		id = CRIA_SERV5_EVT_FL_FBF_ERR;
		break;
	case EDAC:
		break;
	case CORE1553BRM:
		id = CRIA_SERV5_EVT_1553_ERR_H;
		break;
	case GRSPW2:
		id = CRIA_SERV5_EVT_SPW_ERR_H;
		break;
	case ERRLOG:
		id = CRIA_SERV5_EVT_FL_EL_ERR;
		break;
	case SYNC:
		break;
	}

	return id;
}


/**
 * @brief event report interface to IASW
 * @param c the error class
 * @param s the erorr severity
 * @param errno an error number
 *
 */

void event_report(enum error_class c, enum error_severity s, uint32_t err)
{
	uint16_t class_id;
	uint16_t event_data[2] = {0, 0};


	event_data[0] = (uint16_t) (err >> 16) & 0xffff;
	event_data[1] = (uint16_t) err & 0xffff;


	switch (s) {
	case NORMAL:
		class_id = event_report_translate_class_norm(c);
		if (class_id)
			CrIaEvtRaise(CRIA_SERV5_EVT_NORM,
				   class_id, event_data, 4);
		break;
	case LOW:
		class_id = event_report_translate_class_low(c);
		if (class_id)
			CrIaEvtRaise(CRIA_SERV5_EVT_ERR_LOW_SEV,
				   class_id, event_data, 4);
		break;
	case MEDIUM:
		class_id = event_report_translate_class_med(c);
		if (class_id)
			CrIaEvtRaise(CRIA_SERV5_EVT_ERR_MED_SEV,
				   class_id, event_data, 4);
		break;
	case HIGH:
		class_id = event_report_translate_class_high(c);
		if (class_id)
			CrIaEvtRaise(CRIA_SERV5_EVT_ERR_HIGH_SEV,
				   class_id, event_data, 4);
		break;
	default:
		break;
	}
}
